import logging
import tempfile
from pathlib import Path
from typing import Optional

from pydantic import BaseSettings, Field

from landingai.storage.data_access import download_file

_LOGGER = logging.getLogger(__name__)


class SnowflakeCredential(BaseSettings):
    """Snowflake API credential. It's used to connect to Snowflake.
    It supports loading from environment variables or .env files.

    The supported name of the environment variables are (case-insensitive):
    - SNOWFLAKE_USER
    - SNOWFLAKE_PASSWORD
    - SNOWFLAKE_ACCOUNT

    Environment variables will always take priority over values loaded from a dotenv file.
    """

    user: str
    password: str
    account: str

    class Config:
        env_file = ".env"
        env_prefix = "SNOWFLAKE_"
        case_sensitive = False


class SnowflakeDBConfig(BaseSettings):
    """Snowflake connection config.
    It supports loading from environment variables or .env files.

    The supported name of the environment variables are (case-insensitive):
    - SNOWFLAKE_WAREHOUSE
    - SNOWFLAKE_DATABASE
    - SNOWFLAKE_SCHEMA

    Environment variables will always take priority over values loaded from a dotenv file.
    """

    warehouse: str
    database: str
    # NOTE: the name "schema" is reserved by pydantic, so we use "snowflake_schema" instead.
    snowflake_schema: str = Field(..., env="SNOWFLAKE_SCHEMA")

    class Config:
        env_file = ".env"
        env_prefix = "SNOWFLAKE_"
        case_sensitive = False


def save_remote_file_to_local(
    remote_filename: str,
    stage_name: str,
    *,
    local_output: Optional[Path] = None,
    credential: Optional[SnowflakeCredential] = None,
    connection_config: Optional[SnowflakeDBConfig] = None,
) -> Path:
    """Save a file stored in Snowflake to local disk.
    If local_output is not provided, a temporary directory will be created and used.
    If credential or connection_config is not provided, it will read from environment variable or .env file instead.
    """
    url = get_snowflake_presigned_url(
        remote_filename,
        stage_name,
        credential=credential,
        connection_config=connection_config,
    )
    if local_output is None:
        local_output = Path(tempfile.mkdtemp())
    file_path = local_output / remote_filename
    file_path.parent.mkdir(parents=True, exist_ok=True)
    download_file(url, file_output_path=file_path)
    _LOGGER.info(f"Saved file {remote_filename} to {file_path}")
    return file_path


def get_snowflake_presigned_url(
    remote_filename: str,
    stage_name: str,
    *,
    credential: Optional[SnowflakeCredential] = None,
    connection_config: Optional[SnowflakeDBConfig] = None,
) -> str:
    """Get a presigned URL for a file stored in Snowflake.
    NOTE: Snowflake returns a valid URL even if the file doesn't exist.
          So the downstream needs to check if the file exists first.
    """
    import snowflake.connector  # type: ignore

    if credential is None:
        credential = SnowflakeCredential()
    if connection_config is None:
        connection_config = SnowflakeDBConfig()

    ctx = snowflake.connector.connect(
        user=credential.user,
        password=credential.password,
        account=credential.account,
        warehouse=connection_config.warehouse,
        database=connection_config.database,
        schema=connection_config.snowflake_schema,
    )
    cur = ctx.cursor()
    exec_res = cur.execute(f"LIST @{stage_name}")
    if exec_res is None:
        raise ValueError(f"Failed to list files in stage: {stage_name}")
    files = exec_res.fetchall()
    _LOGGER.debug(f"Files in stage {stage_name}: {files}")
    exec_res = cur.execute(
        f"SELECT get_presigned_url(@{stage_name}, '{remote_filename}') as url"
    )
    if exec_res is None:
        raise ValueError(
            f"Failed to get presigned url for file: {remote_filename} in stage: {stage_name}"
        )
    result = exec_res.fetchall()
    if len(result) == 0 or len(result[0]) == 0:
        raise FileNotFoundError(
            f"File ({remote_filename}) not found in stage {stage_name}. Please double check the file exists in the expected location, stage: {stage_name}, db config: {connection_config}."
        )
    result_url: str = result[0][0]
    _LOGGER.info(f"Result url: {result_url}")
    return result_url



import math
from collections import defaultdict
from itertools import groupby
from typing import Dict, List, Sequence, Tuple, Union, cast

import numpy as np
import PIL.Image
from PIL.Image import Image

from landingai.common import (
    ClassificationPrediction,
    ObjectDetectionPrediction,
    Prediction,
    SegmentationPrediction,
)


def rescale_bboxes_by_image_size(
    predictions: Sequence[ObjectDetectionPrediction],
    from_image: Image,
    to_image: Image,
) -> Sequence[ObjectDetectionPrediction]:
    """Rescale the bounding boxes of each ObjectDetectionPrediction in a list based on the size of the reference images.
    NOTE: this operation is NOT in-place. The input predictions will remain the same.

    Parameters
    ----------
    predictions: a list of ObjectDetectionPrediction to be rescaled.
    from_image: the image that serves the denominator of the scale factor. The input bboxes is at the same scale of this image.
    to_image: the image that serves the numerator of the scale factor. The ouput bboxes will be at the same scale of this image.

    Returns
    -------
    A list of ObjectDetectionPrediction where the bboxes has been rescaled.
    """
    scale_factor = (
        to_image.size[1] / from_image.size[1],
        to_image.size[0] / from_image.size[0],
    )
    return rescale_bboxes(predictions, scale_factor)


def rescale_bboxes(
    predictions: Sequence[ObjectDetectionPrediction],
    scale_factor: Union[Tuple[float, float], float],
) -> Sequence[ObjectDetectionPrediction]:
    """Rescale the bounding boxes of each ObjectDetectionPrediction in a list based on the scale factor.
    NOTE: this operation is NOT in-place. The input predictions will remain the same.

    Parameters
    ----------
    predictions: a list of ObjectDetectionPrediction to be rescaled.
    scale_factor: the scale factor that will be applied to the predictions. The scale factors are (height, width) if it's a tuple.

    Returns
    -------
    A list of ObjectDetectionPrediction where the bboxes has been rescaled.
    """
    if isinstance(scale_factor, float):
        scale_factor = (scale_factor, scale_factor)
    assert scale_factor[0] > 0 and scale_factor[1] > 0, "Scale factor must be > 0"
    height_scale, width_scale = scale_factor
    return [
        ObjectDetectionPrediction(
            id=pred.id,
            score=pred.score,
            label_index=pred.label_index,
            label_name=pred.label_name,
            bboxes=(
                math.floor(pred.bboxes[0] * width_scale),  # xmin
                math.floor(pred.bboxes[1] * height_scale),  # ymin
                math.ceil(pred.bboxes[2] * width_scale),  # xmax
                math.ceil(pred.bboxes[3] * height_scale),  # ymax
            ),
        )
        for pred in predictions
    ]


def crop(
    predictions: Sequence[Prediction], image: Union[np.ndarray, Image]
) -> Sequence[Image]:
    """Crop the image based on the bounding boxes in the predictions.

    NOTE: Currently, only ObjectDetectionPrediction is supported. If other types of predictions are passed in, a ValueError will be raised.

    Parameters
    ----------
    predictions: a list of ObjectDetectionPrediction, each of which will be used to crop the image.
    image: the input image to be cropped from.

    Returns
    -------
    A list of cropped images.
    """
    if len(predictions) == 0:
        return []
    if isinstance(image, np.ndarray):
        image = PIL.Image.fromarray(image)
    output = []
    for pred in predictions:
        if not isinstance(pred, ObjectDetectionPrediction):
            raise ValueError(
                f"Only ObjectDetectionPrediction is supported but {type(pred)} is found."
            )
        output.append(image.crop(pred.bboxes))
    return output


def class_map(predictions: Sequence[ClassificationPrediction]) -> Dict[int, str]:
    """Return a map from the predicted class/label index to the class/label name."""
    return {pred.label_index: pred.label_name for pred in predictions}


def class_counts(
    predictions: Sequence[ClassificationPrediction],
) -> Dict[int, Tuple[int, str]]:
    """Compute the distribution of the occurrences of each class.

    Returns
    -------
    A map with the predicted class/label index as the key, and a tuple of
        (the number of occurrence, class/label name) as the value.
        ```
        Example:
            {
                1: (10, "cat"),
                2: (31, "dog"),
            }
        ```
    """
    counts: Dict[int, List[Union[int, str]]] = defaultdict(lambda: [0, ""])
    for pred in predictions:
        counts[pred.label_index][0] = cast(int, counts[pred.label_index][0]) + 1
        counts[pred.label_index][1] = pred.label_name
    return {k: (cast(int, v[0]), cast(str, v[1])) for k, v in counts.items()}


def class_pixel_coverage(
    predictions: Sequence[ClassificationPrediction],
    coverage_type: str = "relative",
) -> Dict[int, Tuple[float, str]]:
    """Compute the pixel coverage of each class.

    Supported prediction types are:
    - SegmentationPrediction
    - ObjectDetectionPrediction

    It supports two ways to compute the coverage:
    - "absolute"
    - "relative"
    See the documentation of the coverage_type for more details.

    Parameters
    ----------
    predictions: a list of predictions. It could come from one or multiple images.
    coverage_type: "absolute" or "relative".
            - Absolute: The number of pixels of each predicted class.
            - Relative: The percentage of pixels that are predicted as the class
               over the sum total number of pixels of every mask. The only project type that supports "relative" is "SegmentationPrediction".


    Returns
    -------
    A map with the predicted class/label index as the key, and a tuple of
        (the coverage, class/label name) as the value.
        ```
        Example (coverage_type="absolute"):
            {
                0: (23512, "blue"),
                1: (1230, "green"),
                2: (0, "pink"),
            }
        ```
    """
    assert isinstance(
        predictions[0], SegmentationPrediction
    ), "Only support SegmentationPrediction for now."
    predictions = cast(List[SegmentationPrediction], predictions)
    return segmentation_class_pixel_coverage(predictions, coverage_type)


def od_class_pixel_coverage(
    predictions: Sequence[ObjectDetectionPrediction],
    coverage_type: str = "relative",
) -> Dict[int, Tuple[float, str]]:
    raise NotImplementedError()


def segmentation_class_pixel_coverage(
    predictions: Sequence[SegmentationPrediction],
    coverage_type: str = "relative",
) -> Dict[int, Tuple[float, str]]:
    """Compute the pixel coverage of each class.
    The coverage is defined as the percentage of pixels that are predicted as the class
    over the sum total number of pixels of every mask.
    If the predictions are from multiple images, the coverage is the average coverage across all images.

    Parameters
    ----------
    predictions: A list of segmentation predictions. It could come from one or multiple images.

    Returns
    -------
    A map with the predicted class/label index as the key, and a tuple of
        (the coverage percentage, class/label name) as the value.
        Note: The sum of the coverage percentage over all classes is not guaranteed
        to be 1.
        ```
        Example (coverage_type="relative"):
            {
                0: (0.15, "blue"),
                1: (0.31, "green"),
                2: (0.07, "pink"),
            }
        ```
    """
    pixel_coverages = []
    label_map = {}
    for pred in predictions:
        coverage = cast(float, pred.num_predicted_pixels)
        if coverage_type == "relative":
            coverage /= math.prod(pred.mask_shape)
        pixel_coverages.append((pred.label_index, coverage))
        label_map[pred.label_index] = pred.label_name

    sorted(pixel_coverages, key=lambda x: x[0])
    coverage_by_label: Dict[int, Tuple[float, str]] = {}
    for label_index, group in groupby(pixel_coverages, key=lambda x: x[0]):
        cov_vals = [item[1] for item in list(group)]
        avg_coverage = sum(cov_vals) / len(cov_vals) if len(cov_vals) > 0 else 0
        coverage_by_label[label_index] = (avg_coverage, label_map[label_index])
    return coverage_by_label



from typing import Dict, Sequence, cast

from landingai.pipeline.frameset import FrameSet
from landingai.postprocess import class_counts
from landingai.common import ClassificationPrediction


def get_class_counts(
    frs: FrameSet, add_id_to_classname: bool = False
) -> Dict[str, int]:
    """This method returns the number of occurrences of each detected class in the FrameSet.

    Parameters
    ----------
    add_id_to_classname : bool, optional
        By default, detections with the same class names and different defect
        id will be counted as the same. Set to True if you want to count them
        separately

    Returns
    -------
    Dict[str, int]
        A dictionary with the counts
        ```
        Example:
            {
                "cat": 10,
                "dog": 3
            }
        ```
    """
    counts = {}
    for frame in frs.frames:
        # Here is a sample return from class_counts: {1: (3, 'Heart'), 3: (3, 'Club'), 4: (3, 'Spade'), 2: (3, 'Diamond')}
        if frame.predictions._inner_type == "OcrPrediction":
            raise TypeError("Can't count classes for OcrPredictor")
        predictions = cast(Sequence[ClassificationPrediction], frame.predictions)
        for k, v in class_counts(predictions).items():
            if add_id_to_classname:  # This is useful if class names are not unique
                class_name = f"{v[1]}_{k}"
            else:
                class_name = v[1]
            if class_name not in counts:
                counts[class_name] = v[0]
            else:
                counts[class_name] += v[0]
    return counts
